//////////////////GLOBALS//////////////////////
#include "SerialCommand.h"
#include "EEPROMAnything.h"
#define Baudrate 115200

//Sensor + PID
SerialCommand sCmd(Serial1);
unsigned long cycleTime = 1000000;
unsigned long previous;
const int sensor[] = {A11, A10, A9, A0, A1, A2, A3, A4};
int normalisatie[8]; //Array for normalisation values
float dalpuntdebug; 
int calculationTime;
bool start = LOW;
float iTerm = 0;
int lasterror;

//Motor A
const byte Ain1 = 5;
const byte Ain2 = 4;
const int PWMA = 3;

//Motor B
const byte Bin1 = 7;
const byte Bin2 = 8;
const int PWMB = 11;
const byte notStandby = 6;


//////////////////STRUCT//////////////////////
//This struct gets saved into EEPROM memory
struct RobotSettings
{
  unsigned long cycleTime;
  int power;
  float Kp;
  float Ki;
  float Kd;
  float diff;
  int black[8];
  int white[8];
} BotSettings;




//////////////////SETUP//////////////////////
void setup()
{
  //Sensor + PID
  Serial.begin(Baudrate);
  Serial1.begin(Baudrate);
  sCmd.addCommand("set", onSet);
  sCmd.addCommand("cali", onCalibrate);
  sCmd.addCommand("debug", onDebug);
  sCmd.addCommand("start", onStart);
  sCmd.addCommand("stop", onStop);
  sCmd.setDefaultHandler(onUnknownCommand);

  //EEPROM
  EEPROM_readAnything(0, BotSettings); //All EEPROM data gets sent to adress 0

  //Motor Driver
  pinMode(Ain1, OUTPUT);
  pinMode(Ain2, OUTPUT);
  pinMode(PWMA, OUTPUT);
  pinMode(Bin1, OUTPUT);
  pinMode(Bin2, OUTPUT);
  pinMode(PWMB, OUTPUT);
  pinMode(notStandby, OUTPUT);
  digitalWrite(notStandby, HIGH); //Enable Motor Driver

}





///////////////////LOOP///////////////////////

void loop()
{
  sCmd.readSerial();
  unsigned long current = micros();

  //Code to adapt cycletime in the loop
  if ((current - previous) > BotSettings.cycleTime)
  {
    previous = current;


    //Normalisation + PID code
    //"normalisatie" means normalization in Dutch
    for (int i = 0; i < 8; i++)
    {
      normalisatie[i] = map(analogRead(sensor[i]), BotSettings.black[i], BotSettings.white[i], 0, 1000);
    }




    //Code to find the blackest sensor
    int index = 0;

    for (int i = 1; i < 8; i++)
    {
      if (normalisatie[i] < normalisatie[index])
      {
        index = i;
      }
    }


    if (normalisatie[index] > 600) start = false; //When the robot does not detect a black line, stop moving


    //Define 1 numerical value for QTR-8A Sensor
    float dalpunt = 0;
    //Define left and right margins for sensor: Left -> -30 ; Right -> 30
    if (index == 0)
    {
      dalpunt = -30;
    }
    else if (index == 7)
    {
      dalpunt = 30;
    }
    else
    {

      int sensNul   = normalisatie[index];
      int sensMin1  = normalisatie[index - 1];
      int sensPlus1 = normalisatie[index + 1];


      float b = sensPlus1 - sensMin1;
      b = b / 2;
      float a = (sensPlus1 - b - sensNul);
      float c = sensNul;

      
      dalpunt = -b / (2 * a);
      dalpunt += index; 
      dalpunt -= 3.5; 

      dalpunt *= 9.525; //On QTR-8A, all sensors are placed at an equal distance of 9,525 mm 
    }

    dalpuntdebug = dalpunt; //Global variable for dalpunt
    //"dalpunt" is the final sensor value, after normalization. It will be used as an input for the PID controller. 





    //P Controller
    float error = -dalpunt;
    float output = error * BotSettings.Kp;


    //I Controller -> output = Ki*.som(error)
    if (start) //Only execute this code when the robot is started
    {
      iTerm = iTerm + (BotSettings.Ki * error); //Error sum * Ki
      iTerm = constrain(iTerm, -510, 510);
      output += iTerm;
    }

    //D Controller -> output = Kd*.(error - lasterror)
    output += BotSettings.Kd * (error - lasterror);
    lasterror = error;





    //Controller Output
    output = constrain(output, -510, 510); //Constraining output according to physical limitations of Motor Driver


    int powerLeft = 0;
    int powerRight = 0;

    
    if (start) if (output >= 0)
      {
        powerLeft = constrain(BotSettings.power + BotSettings.diff * output, -255, 255);
        powerRight = constrain(powerLeft - output, -255, 255);
        powerLeft = powerRight + output;
      }
      else
      {
        powerRight = constrain(BotSettings.power - BotSettings.diff * output, -255, 255);
        powerLeft = constrain(powerRight + output, -255, 255);
        powerRight = powerLeft - output;
      }



    //Code Left Motor
    if (powerLeft >= 0)
    {
      digitalWrite (Ain1, HIGH);
      digitalWrite (Ain2, LOW);

    }
    else if (powerLeft < 0)
    {
      digitalWrite (Ain1, LOW);
      digitalWrite (Ain2, HIGH);

    }
    analogWrite(PWMA, abs(powerLeft));

    //Code Right Motor
    if (powerRight >= 0)
    {
      digitalWrite(Bin1, HIGH);
      digitalWrite(Bin2, LOW);

    }
    else if (powerRight < 0)
    {
      digitalWrite(Bin1, LOW);
      digitalWrite(Bin2, HIGH);

    }
    analogWrite(PWMB, abs(powerRight));


  }

  //Determining calculationTime
  unsigned long difference = micros() - current;
  if (difference > calculationTime) calculationTime = difference;



}



/////////////////FUNCTIONS/////////////////////
void onSet ()
{

  char* parameter = sCmd.next ();
  char* value = sCmd.next ();


  //This functions enables the ability to read commands
  if (strcmp(parameter, "cycle") == 0)
  {
    long newCycleTime = atol(value);
    float ratio = ((float) newCycleTime) - ((float)BotSettings.cycleTime);

    BotSettings.Ki *= ratio; //Ki gets adjusted when a new cycletime gets detected
    BotSettings.Kd /= ratio;
    BotSettings.cycleTime = newCycleTime;
  }

  else if (strcmp(parameter, "power") == 0) //Adjust speed of robot
  {
    BotSettings.power = atol(value);
    Serial1.print("Power Set To ");
    Serial1.println(BotSettings.power);
  }

  else if (strcmp(parameter, "diff") == 0) //Affects intensity of Parameter "Power" depending on the error amount
  {
    BotSettings.diff = atof(value);
    Serial1.print("Diff Set To ");
    Serial1.println(BotSettings.diff);
  }
  else if (strcmp(parameter, "kp") == 0) //Adjust Kp value
  {
    BotSettings.Kp = atof(value);
    Serial1.print("Kp Set To ");
    Serial1.println(BotSettings.Kp);
  }
  else if (strcmp(parameter, "ki") ==  0) //Adjust Ki value
  {
    float cycleTimeInSec = ((float) BotSettings.cycleTime) / 1000000;
    BotSettings.Ki = atof(value) * cycleTimeInSec;

    Serial1.print("Ki Set To");
    Serial1.println(BotSettings.Ki);
  }
  else if (strcmp(parameter, "kd") == 0) //Adjust Kp value
  {
    float cycleTimeInSec = ((float) BotSettings.cycleTime) / 1000000;
    BotSettings.Kd = atof(value) / cycleTimeInSec;

    Serial1.print("Kd Set To");
    Serial1.println(BotSettings.Kd);
  }
  else Serial1.println ("Unknown value for function: Set");


  //This command writes all edited Parameters into EEPROM memory
  EEPROM_writeAnything(0, BotSettings);
}


void onStart ()
{
  start = HIGH;
  iTerm = 0; //Initialise iTerm when robot starts
}

void onStop ()
{
  start = LOW;
}

void onDebug () //Print all Parameter values to Bluetooth Serial Monitor when instruction "debug" is called
{
  Serial1.print("Cali Black: ");
  for (int i = 0; i < 8; i++)
  {
    Serial1.print(BotSettings.black[i]);
    Serial1.print (" ");
  }
  Serial1.println(" ");

  Serial1.print("Cali White: ");
  for (int i = 0; i < 8; i++)
  {
    Serial1.print(BotSettings.white[i]);
    Serial1.print(" ");
  }
  Serial1.println(" ");

  Serial1.print("Norm Values: ");
  for (int i = 0; i < 8; i++)
  {
    Serial1.print(normalisatie[i]);
    Serial1.print(" ");
  }
  Serial1.println("");

  Serial1.print("Power: ");
  Serial1.println(BotSettings.power);

  Serial1.print("Diff: ");
  Serial1.println(BotSettings.diff);

  Serial1.print("Kp: ");
  Serial1.println(BotSettings.Kp);

  float cycleTimeInSec = ((float) BotSettings.cycleTime) / 1000000;
  float Ki = BotSettings.Ki / cycleTimeInSec;
  Serial1.print("Ki: " );
  Serial1.println(BotSettings.Ki);

  float Kd = BotSettings.Kd * cycleTimeInSec;
  Serial1.print("Kd: " );
  Serial1.println(BotSettings.Kd);


  Serial1.print("Cycle Time: ");
  Serial1.println(BotSettings.cycleTime);

  Serial1.print("Calculation Time: ");
  Serial1.println(calculationTime);

  Serial1.print("Position: ");
  Serial1.println(dalpuntdebug);

  Serial1.println(""); 
  Serial1.println("");
}




void onCalibrate()
{
  char* parameter = sCmd.next();

  if (strcmp(parameter, "black") == 0)
  {
    Serial1.print("Start Calibrating Black... ");
    for (int i = 0; i < 8; i++)
    {
      BotSettings.black[i] = analogRead(sensor[i]);
    }
    Serial1.println("Done!");
  }


  if (strcmp(parameter, "white") == 0)
  {
    Serial1.print("Start Calibrating White... ");
    for (int i = 0; i < 8; i++)
    {
      BotSettings.white[i] = analogRead(sensor[i]);
    }
    Serial1.println("Done!");
  }

  //Write all Calibration values to EEPROM Memory
  EEPROM_writeAnything(0, BotSettings);

}



void onUnknownCommand(char* command)
{
  Serial1.print("Unknown Command: \"");
  Serial1.print(command);
  Serial1.println("\"");
}
